import itertools
from typing import List, Optional, Tuple

from astroid import nodes
from pylint.checkers import BaseChecker

from kognitos.bdk.docstring import DocstringParseError, DocstringParser
from kognitos.bdk.errors import SignatureError
from kognitos.bdk.klang.parser import KlangParser
from kognitos.bdk.reflection import (BookProcedureDescriptor,
                                     BookProcedureSignature)
from kognitos.bdk.reflection.factory import BookProcedureFactory

from . import util


def get_all_input_names(node: nodes.FunctionDef | nodes.AsyncFunctionDef, book_procedure: BookProcedureDescriptor) -> List[str]:
    """
    Retrieves all input concept names as snake case, as well as all the function arguments which are not mapped to any input concept.
    """
    concept_map = {p.python_name: [c.noun_phrases[0].to_snake_case() for c in p.concepts] for p in book_procedure.parameter_concept_map}
    input_names = itertools.chain.from_iterable(concept_map.values())
    func_args = [arg.name for arg in node.args.arguments if arg.name != "self" and arg.name not in concept_map]

    return func_args + list(input_names)


class ProcedureChecker(BaseChecker):
    """
    ProcedureChecker

    This class is responsible for checking the documentation and type annotations of procedures.
    """

    name = "kognitos-procedure-checker"
    msgs = {
        "C7401": (  # message id
            # template of displayed message
            "Procedure %s is missing description",
            # message symbol
            "procedure-missing-description",
            # message description
            "All procedures must have a description",
        ),
        "C7403": (  # message id
            # template of displayed message
            "Procedure %s is missing documentation",
            # message symbol
            "procedure-missing-documentation",
            # message description
            "All procedures must have a documentation block attached to them",
        ),
        "C7404": (  # message id
            # template of displayed message
            "Unable to parse documentation block for procedure %s: %s",
            # message symbol
            "procedure-bad-documentation",
            # message description
            "All procedures must have a correct documentation string attached to them",
        ),
        "C7405": (  # message id
            # template of displayed message
            "Missing type annotations for arg %s",
            # message symbol
            "procedure-argument-missing-type-annotation",
            # message description
            "All procedures must have all of their arguments annotated with type",
        ),
        "C7406": (  # message id
            # template of displayed message
            "Unsupported type for arg %s",
            # message symbol
            "procedure-argument-unsupported-type",
            # message description
            "BCI supports a limited set of types",
        ),
        "C7407": (  # message id
            # template of displayed message
            "Unable to infer type for arg %s",
            # message symbol
            "procedure-argument-cannot-infer-type",
            # message description
            "Internal Error. Unable to infer the type for the argument.",
        ),
        "C7408": (  # message id
            # template of displayed message
            "Missing description for arg %s",
            # message symbol
            "procedure-missing-argument-description",
            # message description
            "All arguments must have a description",
        ),
        "C7409": (  # message id
            # template of displayed message
            "Cannot parse english for procedure %s: %s",
            # message symbol
            "procedure-cannot-parse-english",
            # message description
            "All procedure english sentences must use a well defined format",
        ),
        "C7410": (  # message id
            # template of displayed message
            "Unsupported type for return %s",
            # message symbol
            "procedure-return-unsupported-type",
            # message description
            "BCI supports a limited set of types",
        ),
        "C7411": (  # message id
            # template of displayed message
            "Unable to infer type for return %s",
            # message symbol
            "procedure-return-cannot-infer-type",
            # message description
            "Internal Error. Unable to infer the type for the return type.",
        ),
        "C7412": (  # message id
            # template of displayed message
            "Missing english sentence for procedure %s",
            # message symbol
            "procedure-missing-english",
            # message description
            "All @procedure must include the english signature",
        ),
        "C7413": (  # message id
            # template of displayed message
            "Missing description for return of method %s",
            # message symbol
            "procedure-missing-return-description",
            # message description
            "All returns must have a description",
        ),
        "C7414": (  # message id
            # template of displayed message
            "The method %s returns something, but it isn't declared on the english sentence",
            # message symbol
            "procedure-missing-return",
            # message description
            "All returns must be explicitly declared on the english sentence",
        ),
        "C7415": (  # message id
            # template of displayed message
            "The method %s has more than one filter expression",
            # message symbol
            "procedure-invalid-filter-expression-count",
            # message description
            "A method can have at most one filter expression",
        ),
        "C7416": (  # message id
            # template of displayed message
            "The method %s must return a List of values",
            # message symbol
            "procedure-with-filter-expression-must-return-list",
            # message description
            "A method with a filter expression must return a List of values",
        ),
        "C7417": (  # message id
            # template of displayed message
            "The method argument %s must be of type Optional[int]",
            # message symbol
            "procedure-with-pagination-type-mismatch",
            # message description
            "A method which supports pagination must use Optional[int] for the offset and limit parameters",
        ),
        "C7418": (  # message id
            # template of displayed message
            "The method %s is missing pagination parameters",
            # message symbol
            "procedure-missing-pagination-parameters",
            # message description
            "A method which supports pagination must include both offset and limit parameters",
        ),
        "C7419": (  # message id
            # template of displayed message
            "The method %s must return a List[T] or Optional[List[T]]",
            # message symbol
            "procedure-with-pagination-return-type-mismatch",
            # message description
            "A method which supports pagination must return a List[T] or Optional[List[T]]",
        ),
        "C7420": (  # message id
            "The method %s has an PoS argument (%s) which is a Tuple, but the length of the Tuple types does not match the length of the English signature PoS",
            # message symbol
            "procedure-pos-argument-tuple-length-mismatch",
            # message description
            "The length of the Tuple does not match the length of the English signature",
        ),
        "C7421": (  # message id
            "The method '%s' has a description for argument '%s' which is not in the English signature",
            # message symbol
            "procedure-bad-argument-description",
            # message description
            "The argument in the description is not part of the English signature",
        ),
    }

    @classmethod
    def is_procedure(cls, node: nodes.FunctionDef):
        return bool(util.get_decorator_by_name(node, "kognitos.bdk.decorators.procedure_decorator.procedure"))

    @classmethod
    def procedure_english(cls, node: nodes.FunctionDef) -> Optional[str]:
        decorator = util.get_decorator_by_name(node, "kognitos.bdk.decorators.procedure_decorator.procedure")

        if decorator and hasattr(decorator, "args") and len(decorator.args) > 0:
            return next(decorator.args[0].infer()).value

        return None

    def visit_functiondef(self, node: nodes.FunctionDef | nodes.AsyncFunctionDef) -> None:
        if ProcedureChecker.is_procedure(node):
            # check if the english of the procedure is correct
            english_sentence = ProcedureChecker.procedure_english(node)
            if not english_sentence:
                self.add_message("procedure-missing-english", node=node, args=node.name)
                return

            try:
                english_signature = BookProcedureSignature.from_parser_signature(KlangParser.parse_signature(english_sentence))
            except SyntaxError as error:
                self.add_message("procedure-cannot-parse-english", node=node, args=[node.name, str(error)])
                return

            # verify parameters
            process_pagination_parameters(self, node)
            message_count, filter_expressions = process_input_parameters(self, english_signature, node)

            if filter_expressions and (not node.returns or not util.is_list_type(node.returns)):
                self.add_message("procedure-with-filter-expression-must-return-list", args=node.repr_name(), node=node)
            elif node.returns:
                invalid_return_nodes = util.get_invalid_type_nodes_for_procedure_outputs(node.returns)

                for n in invalid_return_nodes:
                    self.add_message("procedure-return-unsupported-type", args=n.repr_name(), node=n)
                    message_count = message_count + 1

            if message_count > 0:
                return

            # extract signature
            python_signature = util.create_signature(node)

            # parse docstring
            if not node.doc_node:
                self.add_message("procedure-missing-documentation", node=node)
                return

            try:
                docstring = DocstringParser.parse(node.doc_node.value)
            except DocstringParseError as error:
                self.add_message("procedure-bad-documentation", node=node.doc_node, args=[node.name, str(error)])
                return

            # create procedure
            try:
                book_procedure = BookProcedureFactory.create(node.name, english_signature, python_signature, docstring, override_connection_required=False)
            except SignatureError as error:
                if util.is_tuple_arguments_mismatch_error(error):
                    self.add_message(
                        "procedure-missing-return",
                        args=node.repr_name(),
                        node=node,
                    )
                return

            if not book_procedure.short_description and not book_procedure.long_description:
                self.add_message(
                    "procedure-missing-description",
                    args=node.repr_name(),
                    node=node.doc_node,
                )

            for parameter in book_procedure.parameter_concept_map:
                reserved_noun_phrases = ["limit", "offset"]
                argument = [arg for arg in node.args.arguments if arg.name == parameter.python_name][0]
                for concept in parameter.concepts:
                    if not concept.description and parameter.python_name not in reserved_noun_phrases:
                        self.add_message(
                            "procedure-missing-argument-description",
                            args=parameter.python_name,
                            node=argument,
                        )

            if book_procedure.outputs:
                for output in book_procedure.outputs:
                    if not output.description:
                        self.add_message(
                            "procedure-missing-return-description",
                            args=node.repr_name(),
                            node=node,
                        )

                outputs_length = len(book_procedure.outputs)
                if (outputs_length in [0, 1]) and util.is_tuple_type(node.returns):
                    self.add_message(
                        "procedure-missing-return",
                        args=node.repr_name(),
                        node=node,
                    )
                elif outputs_length > 1 and not util.is_tuple_type(node.returns):
                    self.add_message(
                        "procedure-missing-return",
                        args=node.repr_name(),
                        node=node,
                    )

            function_inputs = get_all_input_names(node, book_procedure)
            docstring_inputs = [noun_phrase.to_snake_case() for input_concept in docstring.input_concepts for noun_phrase in getattr(input_concept, "noun_phrases", [])] + [
                param.name for param in docstring.params
            ]  # Both input concepts and arguments

            extra_doc_concepts = sorted(set(docstring_inputs) - set(function_inputs))
            for extra_doc_concept in extra_doc_concepts:
                self.add_message("procedure-bad-argument-description", args=(node.repr_name(), extra_doc_concept), node=node)

    def visit_asyncfunctiondef(self, node: nodes.AsyncFunctionDef) -> None:
        self.visit_functiondef(node)


def process_pagination_parameters(procedure_checker: ProcedureChecker, node: nodes.FunctionDef | nodes.AsyncFunctionDef):

    pagination_arg_names = ["offset", "limit"]
    args_and_annotations = [(arg, node.args.annotations[idx]) for idx, arg in enumerate(node.args.arguments) if arg.name in pagination_arg_names]

    if not args_and_annotations:
        return

    if len(args_and_annotations) != len(pagination_arg_names):
        present_arg_names = [arg.name for (arg, _) in args_and_annotations]

        for name in pagination_arg_names:
            if name not in present_arg_names:
                procedure_checker.add_message("procedure-missing-pagination-parameters", args=name, node=node)

    for arg, annotation in args_and_annotations:
        # All pagination parameters must be of type Optional[int]
        if not (util.is_int_type(annotation) and util.is_optional_type(annotation)):
            procedure_checker.add_message("procedure-with-pagination-type-mismatch", args=arg.name, node=node)

    if not util.is_list_type(node.returns):
        procedure_checker.add_message("procedure-with-pagination-return-type-mismatch", node=node)


def process_input_parameters(procedure_checker: ProcedureChecker, english_signature: BookProcedureSignature, node: nodes.FunctionDef | nodes.AsyncFunctionDef) -> Tuple[int, int]:
    """
    Processes the input parameters of a procedure and returns the number of messages and the number of filter expressions.
    """
    message_count = 0
    filter_expressions_count = 0

    for idx, argument in enumerate(node.args.arguments):
        arg_name = argument.repr_name()
        if arg_name == "self":
            continue

        annotation = node.args.annotations[idx]
        if annotation is None:
            procedure_checker.add_message(
                "procedure-argument-missing-type-annotation",
                args=arg_name,
                node=argument,
            )
            message_count = message_count + 1
        else:
            is_filter_expression = util.is_filter_expression_type(annotation)
            if is_filter_expression:
                filter_expressions_count += 1

            if is_filter_expression and filter_expressions_count > 1:
                procedure_checker.add_message(
                    "procedure-invalid-filter-expression-count",
                    args=arg_name,
                    node=argument,
                )

            if not is_filter_expression:
                # NOTE: This is a special scenario where we allow top-level Tuple input type
                # if we're targetting the PoS object or target.
                if arg_name in ("object", "target") and util.is_tuple_type(annotation):
                    annotation_nodes = annotation.slice.elts

                    if arg_name == "object":
                        input_length_matches = len(annotation_nodes) == len(english_signature.object or [])
                    else:
                        input_length_matches = len(annotation_nodes) == len(english_signature.target or [])

                    if not input_length_matches:
                        procedure_checker.add_message(
                            "procedure-pos-argument-tuple-length-mismatch",
                            args=arg_name,
                            node=argument,
                        )
                        continue

                    for ann_node in annotation_nodes:
                        invalid_nodes = util.get_invalid_type_nodes(ann_node, depth=0, seen=None)
                        if any(invalid_nodes):
                            procedure_checker.add_message(
                                "procedure-argument-unsupported-type",
                                args=arg_name,
                                node=ann_node,
                            )
                            message_count = message_count + 1

                else:
                    invalid_type_nodes = util.get_invalid_type_nodes(annotation, depth=0, seen=None)

                    if any(invalid_type_nodes):
                        procedure_checker.add_message(
                            "procedure-argument-unsupported-type",
                            args=(node.repr_name(), arg_name),
                            node=argument,
                        )
                        message_count = message_count + 1

    return message_count, filter_expressions_count
